/* $Id: 546x.c,v 1.11 1999/09/20 22:38:09 ortalo Exp $
******************************************************************************

   KGI - Cirrus Logic Laguna 546x chipset driver

   Copyright (C) 1997 Chris Lattner	[sabre@magic.hurrah.com]
   Copyright (C) 1997 Rodolphe Ortalo	[ortalo@laas.fr]



   I have added some code to the verify_546x() function, the function doesn't
   seem to work for me. If you want to use the old PCI Device Detection, just
   make a #define USE_OLD_PCI_DETECTION.
                                  <karsten.scheibler@bigfoot.de>, 1999-AUG-29
   
   

   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   THE AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

******************************************************************************
*/

#undef USE_OLD_PCI_DETECTION

#include <kgi/maintainers.h>
#define	MAINTAINER	Rodolphe_Ortalo
#define DRIVER_NAME	"Cirrus Logic CL-GD546x (Laguna) chipset driver"
#define DRIVER_REV	"$Revision: 1.11 $"

/*
 * With kgicon we need to use VGA IO regions wildly
 */
#undef CLAIM_VGA_IO_REGION

/**************************************
**          Include files            **
**************************************/
#include <kgi/kgi.h>
#include <kgi/module.h>
#include "Cirrus/546x.h"

/**************************************
**	Private data structures      **
**************************************/

static struct kgi_chipset chipinfo = {
	"Cirrus Logic",		/* manufacturer		*/
	"CL-GD546x (Laguna)",	/* model		*/
	GM_TEXT16 |          	/* supported modes	*/
	GM_8BIT | GM_15BIT | GM_16BIT |
	GM_24BIT | GM_32BIT,
	0,
	{ 2048, 2048 },		/* maximum resolution	*/
	NULL,			/* Private data hook	*/
	0,			/* RAM size (detected)	*/
	{ 10000000, 230000000}	/* DCLK limits		*/
};

struct StateTable ResetState;

kgi_modereq default_kgi_modereq;

struct kgi_mode default_kgi_mode;

/*****************************
**      Memory regions      **
*****************************/

/* VGA IO Region Start */
#ifdef CLAIM_VGA_IO_REGION
static struct io_region vga_io = {
	IO_NULL,	/* VGA IO Region Start (virtual) */
  	0x3C0,		/* VGA IO Region Start (physical) */
	0x20,		/* 32 register (0x3C0-0x3DF) */
	0x3FF,		/* 10 address bits are decoded	*/
	"CL-GD546x (Laguna) VGA I/O region"
};
#endif

/* VGA Video Mem Region Start (Extended to 128k) */
static struct mem_region vga_mem_graph = {
	MEM_NULL,	/* virtual base adress */
	0xA0000,	/* physical base adress */
	0x10000,	/* size of region */
	-1,		/* decoded phys adress lines */
	"CL-GD546x (Laguna) VGA video framebuffer"
};

/* CL-GD546x Linear memory region start (32 meg) */
/*static*/ struct mem_region linFrameBuffer = {
	MEM_NULL,	/* virtual base adress */
	0,		/* physical base adress (to be set) */
	0x2000000,	/* size of region */
	-1,		/* decoded phys adress lines */
	"CL-GD546x (Laguna) linear video framebuffer"
};

/* CL-GD546x Mem mapped IO Region Start */
static struct mem_region  linIOBuffer = {
	MEM_NULL,	/* virtual base adress */
	0,		/* physical base adress (To be set) */
	0x8000,		/* size of region */
	-1,		/* decoded phys adress lines */
	"CL-GD546x (Laguna) Memory-mapped IO region"
};

/* CL-GD546x MMAP interface for linear framebuffer (for graphics) */
static struct kgi_mmio_region graphic_mmio_linear_fb = {
	NULL,			/* Next */
	"CL-GD546x (Laguna) graphic MMIO framebuffer",	/* Name */
	MMIO_FRAMEBUFFER,	/* Type (Linear) */
	MMIO_RDWR | MMIO_SHARED,/* Protection map */

	0,			/* total size of area */
	0,			/* physical size */
	0,			/* physical start */
	0,			/* offset */

	NULL,		/* set offset function */
	NULL		/* private */
};


/***************************
**   Debugging functions  **
***************************/
#ifdef DEBUG
static void printk_mode (struct kgi_mode *mode, char *msg)
{
	#define M(X)	mode->X
	#define R(X)	mode->request.X
	#define T(X)	M(X.width), M(X.blankstart), M(X.syncstart),\
			M(X.syncend), M(X.blankend), M(X.total), M(X.polarity)

	printk(KERN_DEBUG "%s\n",msg);
	printk(KERN_DEBUG "mode = {\n" 
               KERN_DEBUG "   { %i,  /* frames */\n"
               KERN_DEBUG "   { %i, %i},  /* visible */\n",
			  R(frames), R(visible.x),R(visible.y));
	printk(KERN_DEBUG "   { %i, %i},   /* virtual */\n",
                          R(virt.x),R(virt.y));
	printk(KERN_DEBUG "   { %i, %i},   /* size [mm] */\n",
                          R(size.x),R(size.y));
	printk(KERN_DEBUG "   0x%x,\n", R(graphtype));
	printk(KERN_DEBUG "   { %i, %i}},   /* textgrid */\n",
                          R(dpp.x),R(dpp.y));
	printk(KERN_DEBUG "   { %i, %i, %i, %i, %i, %i, %i}, /* x timing */\n",
                          T(x));
	printk(KERN_DEBUG "   { %i, %i, %i, %i, %i, %i, %i}, /* y timing */\n",
                          T(y));
	printk(KERN_DEBUG "   { %i, %i},   /* magnify */\n",
                          M(magnify.x), M(magnify.y));
	printk(KERN_DEBUG "   %x,          /* flags   */\n", M(flags));
	printk(KERN_DEBUG "   %i,          /* dclk    */\n", M(dclk));
	printk(KERN_DEBUG "   { %i, %i},   /* clk     */\n", 
                          M(clk.mul), M(clk.div));
	printk(KERN_DEBUG "   { %i, %i}    /* lclk    */\n"
	       KERN_DEBUG "}\n",
                          M(lclk.mul), M(lclk.div));

	#undef M
	#undef R
	#undef T
}
#endif /* #ifdef DEBUG */


/***************************
**    C-files inclusion   **
***************************/

/* Chipset IO functions */
#include "cl546xio.inc"

/* Text16 functions */
#include "546x_text16.inc"


/**************************************
** Definition of all the PCI devices **
**************************************/

struct PCIDevice_s PCIDevice[] =
{
   	{ PCI_CL_GD5462, "Cirrus Logic CL-GD5462" },
	{ PCI_CL_GD5464, "Cirrus Logic CL-GD5464" },
	{ 0xd0, "Jon's broken Creative card" },
	{0, NULL}
};


/*********************************
**   Initialisation functions   **
*********************************/

/* Detect the presence of a 546x chipset. */
/* NB: Currently, only PCI devices ! */
static int verify_546x(struct kgi_display *dpy,
			struct chipset_private *Private)
{
   	#ifdef USE_OLD_PCI_DETECTION
	int Status = 0, cnt, i, j;
        #else
        struct pci_dev *dev;
        int i;
        #endif
	kgiu16 cmd = 0;
  
	if (!pcibios_present()) {
		ERROR("Sorry, this driver only supports PCI devices\n");
		return -E(CHIPSET,UNKNOWN);
	}
        #ifdef USE_OLD_PCI_DETECTION
	cnt = 0; /* For initial testing, assumes only one device */
  
	/*	First, figure out if there is a CL-GD546x on the PCI bus.
	*/
	for (i = 0; (PCIDevice[i].string != NULL); i++) {
		j = 0;	/* Start with the first card */
		do {
			Status = pcibios_find_device(PCI_VENDOR_ID_CIRRUS,
                                   PCIDevice[i].id, j,
                                   &(Private->PCIbus),
                                   &(Private->PCIdevfn));
			j++;
 			if (Status == PCIBIOS_SUCCESSFUL) {
				cnt--;
			}
		} while (Status == PCIBIOS_SUCCESSFUL && cnt >= 0);
	}

	if (Status != PCIBIOS_SUCCESSFUL) {
        #else
        for (i = 0; (PCIDevice[i].string != NULL); i++)
           	{
                dev = NULL;                
                dev = pci_find_device(PCI_VENDOR_ID_CIRRUS, PCIDevice[i].id, dev);
                if (dev != NULL) break;
                }
        if (dev != NULL)
           	{
                Private->PCIbus   = dev->bus->number;
                Private->PCIdevfn = dev->devfn;
                }
        else {
        #endif
		ERROR("Failed to find any CL-GD546x (Laguna) on the PCI bus!\n");
		ERROR("  Contact "MAINTAINER" with the contents\n");
		ERROR("  of /proc/pci (if you really have one ;-).\n");
		return -E(CHIPSET,UNKNOWN);
	}

	DEBUG1("Found a CL-GD546x on bus: %d, device ID: %d\n",
		Private->PCIbus, Private->PCIdevfn);

	/*	Enable: access to memory, io ports, and bus mastering.
	*/
	pcibios_read_config_word(Private->PCIbus, Private->PCIdevfn,
				 PCI_COMMAND, &cmd);
	pcibios_write_config_word(Private->PCIbus, Private->PCIdevfn,
		PCI_COMMAND, cmd | PCI_COMMAND_IO | PCI_COMMAND_MEMORY
			| PCI_COMMAND_MASTER);

	/*	Determine Linear framebuffer access.
	*/
	pcibios_read_config_dword(Private->PCIbus, Private->PCIdevfn,
		PCI_BASE_ADDRESS_1, &linFrameBuffer.base); 
	linFrameBuffer.base &= PCI_BASE_ADDRESS_MEM_MASK;
	if (linFrameBuffer.base == 0) {
		ERROR("CL-GD546x invalid linear framebuffer address." );
		return -E(CHIPSET, UNKNOWN);
	}
	DEBUG1("Linear framebuffer address: %x\n", linFrameBuffer.base);

	/*	Determine memory mapped IO access.
	*/
	pcibios_read_config_dword(Private->PCIbus, Private->PCIdevfn,
		PCI_BASE_ADDRESS_0, &linIOBuffer.base); 
	linIOBuffer.base &= PCI_BASE_ADDRESS_MEM_MASK;
	if (linIOBuffer.base == 0) {
		ERROR("CL-GD546x invalid address for memory mapped IO buffer.");
		return -E(CHIPSET, UNKNOWN);
	}
	DEBUG1("Mem-mapped IO base address: %x", linIOBuffer.base);

	/* Check memory regions
	*/
#ifdef CLAIM_VGA_IO_REGION /* Used for /dev/fb & kgicon */
	if (io_check_region(&vga_io)) {
		ERROR("CL-GD546x VGA IO region NOT available !");
		return -EBUSY;
	}
#endif
	if (mem_check_region(&vga_mem_graph)) {
		ERROR("CL-GD546x VGA framebuffer region NOT available !");
		return -EBUSY;
	}
	if (mem_check_region(&linFrameBuffer)) {
		ERROR("CL-GD546x linear framebuffer region NOT available !");
		return -EBUSY;
	}
	if (mem_check_region(&linIOBuffer)) {
		ERROR("CL-GD546x memory mapped IO region NOT available !");
		return -EBUSY;
	}

	/*	Here we know we can get all of the resources to start up 
	**	the chip. Grab them.
	*/
#if CLAIM_VGA_IO_REGION /* Used for /dev/fb & kgicon */
0 /* Test for /dev/fb */
	io_claim_region(&vga_io);
#endif
	mem_claim_region(&vga_mem_graph);
	mem_claim_region(&linFrameBuffer);
	mem_claim_region(&linIOBuffer);

	/*	Read the RamSize from the Scratch register
	*/
	chipinfo.ramsize = ((CLIO_in8(SR14_SCRATCH_2) & 7) + 1) * 1024*1024;
	DEBUG1("Chipset RAM size = %d bytes", chipinfo.ramsize);

	/*	Setup the mmio region for the graphic framebuffer
	*/
	graphic_mmio_linear_fb.size   = chipinfo.ramsize;
	graphic_mmio_linear_fb.psize  = chipinfo.ramsize;
	graphic_mmio_linear_fb.pstart = linFrameBuffer.base;

	return EOK;
}

/*
** Tiling constraints definition table
*/
typedef struct {
	kgi_uint pitch;
	kgi_uint n_tiles;
	kgi_uint wide_tiles;
} laguna_tiles_per_line;

static laguna_tiles_per_line laguna_tiled_pitch_table[] = {
/* We should use small tiles, so we put them in the beginning of the
   table. (Reorder if needed for real performance.) */
	/*  pitch, no_tiles, wide_tiles */
	{     640,        5,          0 },
	{    1024,        8,          0 },
	{    1280,       10,          0 },
	{    1664,       13,          0 },
	{    2048,       16,          0 },
	{    2560,       20,          0 },
	{    3328,       26,          0 },
	{    4096,       32,          0 },
	{    1280,        5,          1 },
	{    2048,        8,          1 },
	{    2560,       10,          1 },
	{    3328,       13,          1 },
	{    4096,       16,          1 },
	{    5120,       20,          1 },
	{    6656,       26,          1 },
	{    8192,       32,          1 },
	{       0,        0,          0 }	/* End flag (on pitch) */
};

static int find_laguna_correct_pitch(kgi_uint proposedpitch,
	kgi_uint* goodpitch, kgi_uint* ntiles, kgi_uint* wide)
{
	int i;

	for (i = 0; laguna_tiled_pitch_table[i].pitch != 0; i++)
	{
		if ( laguna_tiled_pitch_table[i].pitch >= proposedpitch )
		{
			(*goodpitch) = laguna_tiled_pitch_table[i].pitch;
			(*ntiles) = laguna_tiled_pitch_table[i].n_tiles;
			(*wide) = laguna_tiled_pitch_table[i].wide_tiles;
			DEBUG2("goodpitch=%d ntiles=%d wide=%d", (*goodpitch),
				(*ntiles), (*wide) );
			return 0; /* Incorrect output */ 
		}
	}
	DEBUG1("Unable to find tiling parms for pitch=%d",proposedpitch);
	return -1; /* Error */
}

static void adjust_timing(struct kgi_mode *mode)
{
	#define ADJUST(FIELD) mode->x.FIELD -= (mode->x.FIELD % CharWidth)
  
	register int CharWidth = (mode->request.graphtype == KGIGT_TEXT16)
                          ? mode->request.dpp.x : 8; 
  
	ADJUST(width);
	ADJUST(blankstart);
	ADJUST(syncstart);
	ADJUST(syncend);
	ADJUST(blankend);
  
	if (mode->x.total % CharWidth) {
		mode->x.total += CharWidth - (mode->x.total % CharWidth);
	}
    
	if (((mode->x.syncend  / CharWidth) & 0x3F) == 
		((mode->x.syncstart / CharWidth) & 0x3F)) {

		mode->x.syncend -= CharWidth;
	}
  
	if (((mode->x.blankend  / CharWidth) & 0x7F) == 
		((mode->x.blankstart / CharWidth) & 0x7F)) {

		mode->x.blankend -= CharWidth;
	}

	if ((mode->y.syncend & 0xF) == (mode->y.syncstart & 0xF)) {
		mode->y.syncend--;
	}

	if ((mode->y.blankend & 0xFF) == (mode->y.blankstart & 0xFF)) {
		mode->y.blankend--;
	}

	#undef ADJUST
}

/*
 * Takes a mode and a bits-per-pixel value
 */
static void chipset_setup_mode(struct kgi_mode *mode)
{
	int bpp=0;

	NOTICE2("Entering chipset_setup_mode");

	switch (mode->request.graphtype) {
		case KGIGT_TEXT16:
			#warning base or baseptr, TEXT_LEN or area size ? -- ortalo
			mode->fb     = vga_mem_graph.baseptr;
			mode->fb_size= vga_mem_graph.size;
			mode->fsize  = mode->request.virt.x
					* mode->request.virt.y * 2;
			mode->mem    = &text16_color_mem;
			mode->io     = &text16_color_io;
			mode->mmio   = NULL;
			return; /* Done setting up text mode... */
		case KGIGT_1BIT:
			bpp=1;
			break;
		case KGIGT_4BIT:
			bpp=4;
			break;
		case KGIGT_8BIT:
			bpp=8;
			break;
		case KGIGT_15BIT:
		case KGIGT_16BIT:
			bpp=16;
			break;
		case KGIGT_24BIT:
			bpp=24;
			break;
		case KGIGT_32BIT:
			bpp=32;
			break;
		default:
			INTERNAL_ERROR;
			return;
	}
	/* By default, setup for graphics modes...
	 */
	#warning base or baseptr, ramsize or area size ? -- ortalo
	mode->fb     = linFrameBuffer.baseptr;
	mode->fb_size = chipinfo.ramsize;
	mode->fsize   = (mode->request.virt.x*mode->request.virt.y*bpp)/8;
	mode->mem     = NULL;
	mode->io      = NULL;

	    /*\
	    |*|        Setup mmap 
            \*/

	      switch (mode->request.graphtype) {
	      case KGIGT_TEXT16:
		  mode->mmio = NULL;	
		break;
	      case KGIGT_4BIT:
		break;
	      case KGIGT_1BIT:
	      case KGIGT_8BIT:
	      case KGIGT_15BIT:
	      case KGIGT_16BIT:
	      case KGIGT_24BIT:
	      case KGIGT_32BIT:
		  graphic_mmio_linear_fb.next=mode->mmio;
		  mode->mmio=&graphic_mmio_linear_fb;
		break;
	      default:
	      }

	NOTICE2("Leaving chipset_setup_mode");
}

static void chipset_setup_dpy(struct kgi_display *dpy,struct kgi_mode *mode)
{
	struct chipset_private *private = KGIM_PRIVATE(dpy,chipset);

	NOTICE2("Entering chipset_setup_dpy");

	private->splitline = -1;
	private->origin = -1;
	private->last_offset = -1;
	private->last_xdot = -1;
	private->last_ydot = -1;

	switch (mode->request.graphtype) {

	case KGIGT_TEXT16:
		dpy->set_font = text16_set_font;
		dpy->set_origin = text16_set_origin;
		dpy->set_split = text16_set_splitline;
		dpy->set_frame = NULL;
		dpy->show_cursor = text16ShowCursor;
		dpy->hide_cursor = text16HideCursor;
		dpy->undo_cursor = text16UndoCursor;
		/* TODO: Activate pointer
		dpy->show_pointer = text16ShowPointer;
		dpy->hide_pointer = text16HidePointer;
		dpy->undo_pointer = text16UndoPointer;
		*/
		dpy->fb.dpy = dpy;
		dpy->fb.size.x = mode->request.virt.x;
		dpy->fb.size.y = mode->request.virt.y;
		dpy->fb.bpp    = MAP_16_BPP;
		dpy->fb.pbuf   = vga_mem_graph.baseptr; /* virtual ? */
		dpy->fb.zbuf   = NULL;
		dpy->fb.ops    = &text16_color_io;
		break;

	default:
		dpy->set_font=NULL;
		dpy->set_origin=graphic_set_origin;
		dpy->set_split=text16_set_splitline;
		dpy->set_frame=NULL;
		dpy->show_cursor=NULL;
		dpy->hide_cursor=NULL;
		dpy->undo_cursor=NULL;
		dpy->show_pointer=NULL;
		dpy->hide_pointer=NULL;
		dpy->undo_pointer=NULL;
		dpy->fb.dpy = dpy;
		dpy->fb.size.x = 0;
		dpy->fb.size.y = 0;
		dpy->fb.bpp    = 0;
		dpy->fb.pbuf   = NULL;
		dpy->fb.zbuf   = NULL;
		dpy->fb.ops    = NULL;
		kgim_accel_set_mode(dpy,mode);
	}
	memcpy(&(dpy->mode),mode,sizeof(struct kgi_mode));

	NOTICE2("Leaving chipset_setup_dpy");
}
/*	GetResetState
** This procedure gets the main register of the card and reconstructs
** a mode that it sets as the default ggi mode for the given display.
** It is used at initialisation time to analyse the mode supposedly
** built by the card's BIOS.
**  THIS CODE STILL IS RATHER ILL-WRITTEN !!! (Furthermore, the mode
**    it gives as a default mode for the display is rather critical..)
** It probably only works for a default mode that is VGA compatible
** (and probably only a text mode).
*/
static void GetResetState(struct kgi_display *dpy)
{
	/* Default parameters probed at init state */ 
	int MagX = 1, MagY = 1;
	int DCLKdiv = 8;
	int char_cell_height = 8;
	int DotClock;
	kgi_graphtype GMode;
	struct kgi_timing tx, ty;

	#define GETREG(REGISTER, SIZE) \
		CLIO_GetStateRegister(&ResetState, REGISTER, SIZE)

	NOTICE2("Entering GetResetState");
	/* Get a whole register range in the ResetState*/
	SaveState(&ResetState);
	/* Try to figure out what is the original default mode (gr./txt) */
	if (ResetState.ATC[AR10_MODE & IOMask] &
		AR10_GRAPHICS_MODE) {

		switch (GETREG(GR5_MODE, 1) & GR5_WRITE_MODE_MASK) {

		case GR5_WRITE_MODE_0:
			GMode = KGIGT_8BIT;	/* 8 Bit Mode */
			break;

		case GR5_WRITE_MODE_1:
			GMode = KGIGT_1BIT;	/* 1 Bit Mode */
			break;

		case GR5_WRITE_MODE_3:
			GMode = KGIGT_4BIT;	/* 4 Bit Mode */
			break;

		case GR5_WRITE_MODE_2:
			INTERNAL_ERROR;		/* Unknown weird mode! */
			break;			/* Shouldn't get here! */
		}
		/* otherwise GMode might be uninitialized! -- ortalo */
		GMode = KGIGT_8BIT;
  	} else {
		GMode = KGIGT_TEXT16;
	}
	/* Get the dot clock divider */
	DCLKdiv = (GETREG(SR1_CLOCKING_MODE, 1) & SR1_8_DOT_CLOCK)
		? 8 : 9;
	/* Get the char cell dimensions	(same usually) */
	char_cell_height = (GETREG(CR9_CHARACTER_CELL_HEIGHT, 1) &
		CR9_CELL_HEIGHT_MASK) + 1;
	
	/* We should care about REGISTER EXTENSIONS !!! -- ortalo 
	** But, in fact, we don't care a lot about the initial state
	** set up by the card BIOS: it is only an emergency mode!
	** (and it may even be replaced by a self-made VGA mode if it
	**  happens to be very reliable).
	*/
	/*	Get the Horizontal Timing
	*/
	tx.width = (GETREG(CR1_HORIZONTAL_DISPLAY_END, 1)+1) * DCLKdiv;
	tx.blankstart = GETREG(CR2_HORIZONTAL_BLANKING_START, 1) * DCLKdiv;
	tx.syncstart = GETREG(CR4_HORIZONTAL_SYNC_START , 1) * DCLKdiv;
	tx.syncend =( ( GETREG(CR5_HORIZONTAL_SYNC_END, 1) 
			& CR5_HOR_SYNC_END_MASK )
		      /*
		      | (GETREG(CR4_HORIZONTAL_SYNC_START , 1) & ~0x1F)
		      */
		    ) * DCLKdiv;
	while (tx.syncend < tx.syncstart)
		tx.syncend += 0x20 * DCLKdiv;
	tx.blankend = ( ((GETREG(CR3_HORIZONTAL_BLANKING_END, 1)
			& CR3_HOR_BLANK_END_MASK)
	  	  | ( (GETREG(CR5_HORIZONTAL_SYNC_END, 1)
			& CR5_HOR_BLANK_END_5)
		      ? 0x20 : 0)
		  ) | (GETREG(CR2_HORIZONTAL_BLANKING_START, 1) & ~0x3F) )
		* DCLKdiv;
	if (tx.blankend < tx.blankstart)
		tx.blankend += 0x40 * DCLKdiv ;

	tx.total      = (GETREG(CR0_HORIZONTAL_TOTAL, 1)+5) * DCLKdiv;
 	#warning polarity should be opposite ? -- ortalo
	tx.polarity   = (GETREG(MISC_INPUT, 1) & MISC_NEG_H_SYNC) != 0;
 
	/*	Get the Vertical Timing (with extensions)
	*/
	ty.width = 1 + (GETREG(CR12_VERTICAL_DISPLAY_END, 1) |
		((GETREG(CR7_OVERFLOW, 1) & CR7_VER_DISPLAY_END_8)
			? 0x100 : 0) |
		((GETREG(CR7_OVERFLOW, 1) & CR7_VER_DISPLAY_END_9)
			? 0x200 : 0));
	ty.blankstart = 1 + ( GETREG(CR15_VERTICAL_BLANK_START, 1)
			| ((GETREG(CR7_OVERFLOW, 1) & CR7_VER_BLANK_START_8)
				? 0x100 : 0)
			| ((GETREG(CR9_CHARACTER_CELL_HEIGHT, 1)
				& CR9_VER_BLANK_START_9) ? 0x200 : 0)
			);
	ty.syncstart = GETREG(CR10_VERTICAL_SYNC_START, 1)
			| ((GETREG(CR7_OVERFLOW, 1) & CR7_VER_SYNC_START_8)
				? 0x100 : 0)
			| ((GETREG(CR7_OVERFLOW, 1) & CR7_VER_SYNC_START_9)
				? 0x200 : 0)
			;
	ty.syncend = ( GETREG(CR11_VERTICAL_SYNC_END, 1)
				& CR11_VER_SYNC_END_MASK )
			| (ty.syncstart & ~0xF);
	if (ty.syncend < ty.syncstart)
		ty.syncend += 0x10;
	ty.blankend = (GETREG(CR16_VERTICAL_BLANK_END, 1)
		| (ty.blankstart & ~0xFF)) + 1;
	if (ty.blankend < ty.blankstart)
		ty.blankend += 0x100;
	ty.total = 2 + ( GETREG(CR6_VERTICAL_TOTAL, 1)
		| ((GETREG(CR7_OVERFLOW, 1) & CR7_VER_TOTAL_8) ? 0x100 : 0)
		| ((GETREG(CR7_OVERFLOW, 1) & CR7_VER_TOTAL_9) ? 0x200 : 0)
		);
 	#warning polarity should be opposite ? -- ortalo
	ty.polarity = (GETREG(MISC_INPUT, 1) & MISC_NEG_V_SYNC) != 0;

	DEBUG2("DCLKdiv=%i tx.width=%i tx.blankstart=%i tx.syncstart=%i tx.syncend=%i\n", DCLKdiv, tx.width, tx.blankstart, tx.syncstart, tx.syncend);
	DEBUG2("tx.blankend=%i", tx.blankend);
	DEBUG2("tx.total=%i tx.polarity=%i", tx.total, tx.polarity);

	DEBUG2("DCLKdiv=%i ty.width=%i ty.blankstart=%i ty.syncstart=%i ty.syncend=%i\n", DCLKdiv, ty.width, ty.blankstart, ty.syncstart, ty.syncend);
	DEBUG2("ty.blankend=%i", ty.blankend);
	DEBUG2("ty.total=%i ty.polarity=%i", ty.total, ty.polarity);

	#warning we have more clock modes for initial probe... -- ortalo
	DotClock= (CLK_in8(dpy,0)) ? 28322000 : 25175000;

	/* Right now, we don't have a default mode */
	/* Set all the mode info that we've just gathered....
	*/
	default_kgi_modereq.frames=1;
/* Is the following right? - MenTaLboY */
	default_kgi_modereq.visible.x=tx.width/DCLKdiv;
	default_kgi_modereq.visible.y=ty.width/char_cell_height;
	default_kgi_modereq.virt.x=tx.width/DCLKdiv;
	default_kgi_modereq.virt.y=ty.width/char_cell_height;
	default_kgi_modereq.size.x=0;
	default_kgi_modereq.size.y=0;
	default_kgi_modereq.graphtype=GMode;
	default_kgi_modereq.dpp.x= DCLKdiv;
	default_kgi_modereq.dpp.y= char_cell_height;
	if (dpy->defmode==NULL)
		dpy->defmode=&default_kgi_modereq;
	else
		(*dpy->defmode)=default_kgi_modereq;

	dpy->mode.request=default_kgi_modereq;

	dpy->mode.x=tx;
	dpy->mode.y=ty;
	dpy->mode.magnify.x=1;
	dpy->mode.magnify.y=1;
	dpy->mode.dclk=DotClock;
	dpy->mode.clk.mul=1;
	dpy->mode.clk.div=1;
	dpy->mode.lclk.mul=1;
	dpy->mode.lclk.div=1;
	dpy->mode.chipset=0;

	default_kgi_mode=dpy->mode;

/* TODO: Check the timing adjustements for this initial mode
	adjust_timing(&default_kgi_mode);
*/
	chipset_setup_mode(&default_kgi_mode);
	chipset_setup_dpy(dpy,&(default_kgi_mode));

	#undef GETREG
	  #ifdef DEBUG
	TRACE2(printk_mode(&default_kgi_mode,"Default KGI mode:"));
	#endif
	NOTICE2("Leaving GetResetState"); 
}

/********************************
**                             **
**  GGI MODULE MAIN FUNCTIONS  **
**                             **
********************************/
kgi_sint kgim_chipset_init(struct kgi_card *card, struct kgi_display *dpy)
{
	struct chipset_private *private;

	NOTICE2("Entering kgim_chipset_init");
	NOTICE(DRIVER_NAME " " DRIVER_REV);

	/* Allocate and sets the private data area for this card's chip */
	if (! (private = kgi_malloc_small(sizeof(*private)))) {
		NOTICE("CL-GD546x: Failed to allocate chipset private data.");
		return -ENOMEM;
	} 
	memset(private, 0, sizeof(*private));
	KGIM_PRIVATE(dpy,chipset) = private;

	/* Fill in private info and verify that we have a 546x */ 
	if (verify_546x(dpy, private) != EOK) {
		/* TODO: Free the private data -- ortalo */
		ERROR("Failed to detect a Cirrus Logic CL-GD546x !");
		return -E(CHIPSET, UNKNOWN);
	} 
	/* We get the default mode for the card directly from
	** the card. TODO: Explore a new design. */
	GetResetState(dpy);

	/* Fill in card info structures */
	strcpy(card->manufact, chipinfo.manufact);
	strcpy(card->model, chipinfo.model);
	card->chipset = &chipinfo;
	card->modes   = chipinfo.modes;
	card->flags   = chipinfo.flags;
	card->maxdots = chipinfo.maxdots;
	card->preset  = default_kgi_mode; /* This mode was probed. */
  
	NOTICE2("Leaving kgim_chipset_init");
	return EOK;
}

void kgim_chipset_done(struct kgi_display *dpy)
{
	/* We restore the original state */
	/* TODO: only VGA registers are restored in fact. We should
	** make sure that the restored mode is clean. This is linked to
	** the correct computation of a default mode... (No?)
	*/
	RestoreState(&ResetState);

	/* Free allocated regions */
#if CLAIM_VGA_IO_REGION /* Used for /dev/fb & kgicon */
	io_free_region(&vga_io);
#endif
	mem_free_region(&vga_mem_graph);
	mem_free_region(&linFrameBuffer);
	mem_free_region(&linIOBuffer);
	kgi_free_small(KGIM_PRIVATE(dpy,chipset));
	NOTICE(DRIVER_NAME " V:" DRIVER_REV " removed.");
}

kgi_sint kgim_chipset_check_mode
(struct kgi_display *dpy, struct kgi_mode *mode,
 enum kgi_tm_cmd *cmd)
{
	kgi_uint bpp = 0;
	kgi_uint byte_pp = 0;
	kgi_uint good_pitch = 0;
	kgi_uint ntiles = 0;
	kgi_uint wide_tile = 0;

	NOTICE2("Entering chipset_check_mode");

	if (mode->flags & FL_MON_INTERLACE) {
		NOTICE("Interlaced modes not supported (yet)...");
		return -E(CHIPSET, FAILED);
	}

	/* Checks depth */  
	switch(mode->request.graphtype) {
	case KGIGT_32BIT:
		bpp = 32;
		byte_pp = 4;
		break;
	case KGIGT_16BIT:
	case KGIGT_15BIT:
		bpp = 16;
		byte_pp = 2;
		break;
	case KGIGT_8BIT:
		bpp = 8;
		byte_pp = 1;
		break;
#if 0	/* Not supported yet */
	case KGIGT_4BIT:
		bpp = 4;
		break;
	case KGIGT_1BIT:
		bpp = 1;
		break;
#endif
	case KGIGT_TEXT16:
		bpp = 0;
		break;
	default: 
		DEBUG1("GraphType %i not supported", mode->request.graphtype);
		return -E(CHIPSET, UNKNOWN);
	}

	#define XVIRTUAL (mode->request.virt.x)
	#define YVIRTUAL (mode->request.virt.y)
	#define XVISIBLE (mode->request.visible.x)
	#define YVISIBLE (mode->request.visible.y)
	#define TEXTX    (mode->request.dpp.x)
	#define TEXTY    (mode->request.dpp.y)
	#define TEXT_LEN 0x10000

	/* Dispatch on command */
	switch(*cmd) { 
	case CMD_PROPOSE:
		NOTICE2("check_mode:CMD_PROPOSE");
		mode->magnify.x = mode->magnify.y = 1;
		if (mode->request.graphtype == KGIGT_TEXT16) {    
			if ((TEXTX < 8) || (TEXTX > 9)) {
				DEBUG1("Invalid char width %i!", TEXTX);
				return -EINVAL;
			}
			if (!(XVIRTUAL || YVIRTUAL)) {
				NOTICE1("Need X or Y virtual!");
				return -EINVAL;
			}
			if (!XVIRTUAL) {
				XVIRTUAL = TEXT_LEN / (YVIRTUAL*2);
				DEBUG2("Setting X Virtual = %i", XVIRTUAL);
			}
			if (!YVIRTUAL) {
				YVIRTUAL = TEXT_LEN / (XVIRTUAL*2);
				DEBUG2("Setting Y Virtual = %i", YVIRTUAL);
			}
			if ( ( 2 * XVIRTUAL * YVIRTUAL ) > TEXT_LEN ) {
				NOTICE1("not enough text memory.");
				return -ENOMEM;
			}

			/* TODO: We should check that visible is less
				than virtual I think ? */
			/* This is not useful: we can do virtual text although
			it is not used a lot... ;-)
			XVIRTUAL = XVISIBLE;
			YVIRTUAL = YVISIBLE;
			*/

		} else {	/* if (mode->request.graphtype == KGIGT_TEXT16) */
			if (!find_laguna_correct_pitch((XVIRTUAL*byte_pp),
				&good_pitch, &ntiles, &wide_tile)) {
				XVIRTUAL = (good_pitch / byte_pp);
			} else {
				DEBUG1("Cannot manage %i horizontal pixels!",
				       XVIRTUAL);
				return -E(CHIPSET, FAILED);
			}
			if ((XVIRTUAL*YVIRTUAL*bpp/8) > chipinfo.ramsize) {
				DEBUG1("Not enough memory for %ix%i@%ibpp",
				       XVIRTUAL, YVIRTUAL, bpp);
				return -ENOMEM;
			}
			mode->request.visible.x &= 0x7FFFFFF8;
			mode->request.visible.y &= 0x7FFFFFF8;
		}
   
		mode->x.width = XVISIBLE * TEXTX;
		mode->y.width = YVISIBLE * TEXTY;

		NOTICE2("Leaving chipset_check_mode");
		return EOK;
      
	case CMD_LOWER:
    
		NOTICE2("check_mode:CMD_LOWER");

		if (mode->dclk < chipinfo.dclk.min) {

			DEBUG1("%i hz DCLK is too low", mode->dclk);
			return -E(CHIPSET,FAILED);
		}
    
		mode->dclk = (mode->dclk > chipinfo.dclk.max)
			? chipinfo.dclk.max : mode->dclk;
		/*
		adjust_timing(mode);
		*/
		NOTICE2("Leaving chipset_check_mode");
		return EOK;

	case CMD_RAISE:

		NOTICE2("check_mode:CMD_RAISE");

		if (mode->dclk > chipinfo.dclk.max) {

			DEBUG1("%i hz DCLK is too high", mode->dclk);
			return -E(CHIPSET,FAILED);
		}
    
		mode->dclk = (mode->dclk < chipinfo.dclk.min)
			? chipinfo.dclk.min : mode->dclk;
		/*
		adjust_timing(mode);
		*/

		NOTICE2("Leaving chipset_check_mode");
		return EOK;

	case CMD_CHECK:
  
		NOTICE2("check_mode:CMD_CHECK");

		if (mode->request.graphtype == KGIGT_TEXT16) {

			mode->flags |= FL_CHIP_COLORTEXT;

			if ( (mode->magnify.x != 1) ||
			     (mode->magnify.y != 1) ||
			     (TEXTX < 8) || (TEXTX > 9) ||
			     (TEXTY < 1) || (TEXTY > 32) ||
			     ((2*(XVIRTUAL/TEXTX)*(YVIRTUAL/TEXTY)) >
					TEXT_LEN) ||
			     (mode->dclk > chipinfo.dclk.max) ||
			     (mode->dclk < chipinfo.dclk.min) )
			{
				return -E(CHIPSET,FAILED);
			}
		} else {
		       if ( ((XVIRTUAL*YVIRTUAL*byte_pp) > chipinfo.ramsize) ||
			     (mode->magnify.x != 1) ||
			     (mode->magnify.y != 1) ||
			     (mode->clk.mul != 1) || (mode->clk.div != 1) ||
			     (mode->dclk < chipinfo.dclk.min) ||
			     (mode->dclk > chipinfo.dclk.max) ||
			     (find_laguna_correct_pitch(XVIRTUAL*byte_pp,
				&good_pitch, &ntiles, &wide_tile)) ||
			     ((XVIRTUAL*byte_pp) != good_pitch) )
			{
				NOTICE1("One check condition not met !");
				return -E(CHIPSET, UNKNOWN);
			}
		}
		#warning adjust timing needed here ? -- ortalo
/*
		adjust_timing(mode);
*/
		NOTICE2("Returning OK for CMD_CHECK");

		chipset_setup_mode(mode);

		NOTICE2("Leaving chipset_check_mode");

		return EOK;
    
	default:
		INTERNAL_ERROR;
		return -E(CHIPSET, FAILED);
	}

	#undef XVIRTUAL
	#undef YVIRTUAL
	#undef XVISIBLE
	#undef YVISIBLE
	#undef TEXTX   
	#undef TEXTY   
	#undef TEXT_LEN 
}

void kgim_chipset_set_mode(struct kgi_display *dpy, struct kgi_mode *mode)
{
#define DIRECT_CARD_PROG
	IS_CRITICAL;
	#warning start from ResetState or a default state ? -- ortalo

#ifndef DIRECT_CARD_PROG
	struct StateTable setup = ResetState;
#endif
	int LogicalWidth;
	int i;
	int DCLKdiv;
	struct chipset_private *private = KGIM_PRIVATE(dpy,chipset);
	kgi_uint n_tiles;
	kgi_uint wide_tile;
	kgi_uint good_pitch;

	NOTICE2("Entering chipset_set_mode");

	/* Useful macros to set the various
	** registers of the card.
	*/
#warning do direct programming of card regs (no restore state) ? -- ortalo
#ifndef DIRECT_CARD_PROG
	#define ORREG8(REGISTER, VALUE)                                    \
			CLIO_SetStateRegister(&setup,(REGISTER),           \
			   CLIO_GetStateRegister(&setup,(REGISTER),1)      \
			   | (VALUE), 1)
	#define ANDREG8(REGISTER, VALUE)                                   \
			CLIO_SetStateRegister(&setup,(REGISTER),           \
			   CLIO_GetStateRegister(&setup,(REGISTER),1)      \
			   & (VALUE), 1)
	#define SETREG8(REGISTER, VALUE)                                   \
			CLIO_SetStateRegister(&setup,(REGISTER),(VALUE),1)
	/* MASK indicates which bits to set (to 0 or to 1) */
	#define SETBITS8(REGISTER, VALUE, MASK)                             \
			CLIO_SetStateRegister(&setup,(REGISTER),            \
				(CLIO_GetStateRegister(&setup, (REGISTER),1)\
				| ((VALUE)&(MASK))) & ((VALUE)|(~(MASK))), 1)
#else
	#define ORREG8(REGISTER, VALUE)                                    \
			CLIO_out8((REGISTER),CLIO_in8((REGISTER)) | (VALUE))
	#define ANDREG8(REGISTER, VALUE)                                   \
			CLIO_out8((REGISTER), CLIO_in8((REGISTER)) & (VALUE))
	#define SETREG8(REGISTER, VALUE)                                   \
			CLIO_out8((REGISTER),(VALUE))
	/* MASK indicates which bits to set (to 0 or to 1) */
	#define SETBITS8(REGISTER, VALUE, MASK)                             \
			CLIO_out8((REGISTER), (CLIO_in8((REGISTER))         \
				| ((VALUE)&(MASK))) & ((VALUE)|(~(MASK))))
	#define ORREG16(REGISTER, VALUE)                                    \
			CLIO_out16((REGISTER),CLIO_in16((REGISTER)) | (VALUE))
	#define ANDREG16(REGISTER, VALUE)                                   \
			CLIO_out16((REGISTER), CLIO_in16((REGISTER)) & (VALUE))
	#define SETREG16(REGISTER, VALUE)                                   \
			CLIO_out16((REGISTER),(VALUE))
	/* MASK indicates which bits to set (to 0 or to 1) */
	#define SETBITS16(REGISTER, VALUE, MASK)                             \
			CLIO_out16((REGISTER), (CLIO_in16((REGISTER))         \
				| ((VALUE)&(MASK))) & ((VALUE)|(~(MASK))))
	#define ORREG32(REGISTER, VALUE)                                    \
			CLIO_out32((REGISTER),CLIO_in32((REGISTER)) | (VALUE))
	#define ANDREG32(REGISTER, VALUE)                                   \
			CLIO_out32((REGISTER), CLIO_in32((REGISTER)) & (VALUE))
	#define SETREG32(REGISTER, VALUE)                                   \
			CLIO_out32((REGISTER),(VALUE))
	/* MASK indicates which bits to set (to 0 or to 1) */
	#define SETBITS32(REGISTER, VALUE, MASK)                             \
			CLIO_out32((REGISTER), (CLIO_in32((REGISTER))         \
				| ((VALUE)&(MASK))) & ((VALUE)|(~(MASK))))
#endif

	#define TEXTX    (mode->request.dpp.x)
	#define TEXTY    (mode->request.dpp.y)

#ifdef DIRECT_CARD_PROG
	ENTER_CRITICAL;

	cl546x_unlock();
#endif
	/* EXTENDED REGISTERS */
	/*	Enable full access to display memory and extensions
	*/ 
	ORREG8(CR1B_EXTENDED_DISPLAY_CONTROL,
		  CR1B_ENABLE_BLANK_END_EXTENSIONS
	        | CR1B_BLANKING_CONTROL /* Normally not useful, but... */
					/* in fact it is MANDATORY ! */
		| CR1B_ENABLE_EXTENDED_ADDRESS_WRAP);

	/*
	**	Find logical screen width (pitch ?)
	**	and graphic mode setup.
	*/
	switch (mode->request.graphtype) {

	case KGIGT_TEXT16:
		LogicalWidth = 2 * mode->request.virt.x;
		LogicalWidth /= 4;
		break;
	case KGIGT_1BIT:
	case KGIGT_4BIT:
		LogicalWidth = mode->request.virt.x >> 4;
		#warning look at the 1BIT and 4BIT modes -- ortalo
		break;

	case KGIGT_8BIT:
		LogicalWidth = mode->request.virt.x >> 3;
		SETREG16(GRAPHICS_VIDEO_FORMAT, GRAPHICS_DEPTH_8BITS | 
			GRAPHICS_FORMAT_PALETTIZED | 
			GRAPHICS_8BITS_DAC_PALETTE);
		break;

	case KGIGT_15BIT:
		LogicalWidth = mode->request.virt.x >> 2;
		SETREG16(GRAPHICS_VIDEO_FORMAT, GRAPHICS_DEPTH_16BITS | 
			GRAPHICS_FORMAT_RGB15);
		break;

	case KGIGT_16BIT:
		LogicalWidth = (mode->request.virt.x >> 3) * 2;
		SETREG16(GRAPHICS_VIDEO_FORMAT, GRAPHICS_DEPTH_16BITS | 
			GRAPHICS_FORMAT_RGB);
		break;

	case KGIGT_32BIT:
		LogicalWidth = mode->request.virt.x >> 1;
		SETREG16(GRAPHICS_VIDEO_FORMAT, GRAPHICS_DEPTH_32BITS | 
			GRAPHICS_FORMAT_RGB);
		break;

	default:
		INTERNAL_ERROR;
		return;
	}

	if (mode->request.graphtype == KGIGT_TEXT16) {
		/* Disable extended display (VGA text modes) */
		ANDREG32(PCIFC_VENDOR_SPECIFIC_CONTROL, ~PCIFC_EXT_DISP);
		/* put to 0 compatible read in CTRC3 (Idem VGA.c) */
		/* CR17 here : various normalized settings */    
		SETREG8(CR17_MODE_CONTROL,
			CR17_ADDRESS_WRAP |
			CR17_TIMING_ENABLE |
			CR17_CGA_SUPPORT |
			CR17_SELECT_ROW_SCAN_COUNTER
			);
		/* Set up cursor size */
		SETREG8(CRA_TEXT_CURSOR_START, 0);
		SETREG8(CRB_TEXT_CURSOR_END, 0x0F);
		/* Disable blinking (...stresses me !) */
		ANDREG8(AR10_MODE, ~AR10_BLINK_ENABLE);
		/* Remove Write protection on CR0-CR7 */
		ANDREG8(CR11_VERTICAL_SYNC_END, ~CR11_WRITE_PROTECT_CR0_7);

		SETREG8(CR14_UNDERLINE_ROW_SCANLINE, CR14_UNDERLINE_MASK);
		/* Enable Plane 0 & 1 */
		SETREG8(SR2_PLANE_MASK, SR2_PLANE_0 | SR2_PLANE_1);
		/* Odd/Even settings */
		ANDREG8(SR4_MEMORY_MODE, ~SR4_ODD_EVEN);
		ORREG8(GR5_MODE, GR5_ODD_EVEN);
		/* Use ext memory */
		ORREG8(SR4_MEMORY_MODE, SR4_EXTENDED_MEMORY);
		/* Text mode */
		ANDREG8(GR6_MISCELLANEOUS, ~GR6_GRAPHICS_MODE);
		/* Chaining */
		ORREG8(GR6_MISCELLANEOUS, GR6_CHAIN_ODD_TO_EVEN);
		/* Memory map location VGA=A000-64k Ext=A000-128k */
		SETBITS8(GR6_MISCELLANEOUS,GR6_MEMORY_MAP_VGA,
				GR6_MEMORY_MAP_MASK);
		/* Try to enable this to take advantage of a 128k
		** vga_mem_graph frame buffer for text ? -- ortalo
		SETBITS8(IOGR6Misc,GR6_MEMORY_MAP_EXTENDED,
			GR6_MEMORY_MAP_MASK);
		*/

		/* Sets several things according
		** to 8/9 character width
		*/
		SETBITS8(SR1_CLOCKING_MODE,
			((TEXTX == 8) ? SR1_8_DOT_CLOCK : 0),
			SR1_8_DOT_CLOCK );
		SETBITS8(AR10_MODE, 
			(TEXTX == 9) ? AR10_LINE_GRAPHICS_ENABLE : 0,
			AR10_LINE_GRAPHICS_ENABLE);
		SETREG8(AR13_PIXEL_PANNING, (TEXTX == 9)
			? AR13_9BIT_NO_SHIFT
			: AR13_8BIT_NO_SHIFT);

         	/* Sets the vertical size of characters
		*/
		SETBITS8(CR9_CHARACTER_CELL_HEIGHT,
			TEXTY-1,
			CR9_CELL_HEIGHT_MASK);

		DCLKdiv = TEXTX;

	} else {	/* if (mode->request.graphtype != KGIGT_TEXT16) */
		#warning check <8BPP graphic mode setting -- ortalo
		DCLKdiv = 8;
		/* Enable extended display (non-VGA graphics modes) */
		ORREG32(PCIFC_VENDOR_SPECIFIC_CONTROL, PCIFC_EXT_DISP);
		/* 8bpp packed pixel (in VGA) */
		ANDREG8(SR7_EXTENDED_SEQ_MODE, ~SR7_8BPP_PACKED_PIXEL); 
		/* Find the good tiling setup */
		i = find_laguna_correct_pitch(LogicalWidth * 8, &good_pitch,
			&n_tiles, &wide_tile);
		if (i) {
			NOTICE1("Unable to find correct tiled laguna pitch for the mode (in set_mode !).");
			INTERNAL_ERROR;
			return;
		}
		if ((LogicalWidth*8) != good_pitch) {
			DEBUG1("Incorrect pitch for tiled mode: current pitch=%d, good pitch=%d, n_tiles=%d, wide=%d", (LogicalWidth*8), good_pitch, n_tiles, wide_tile);
		}
		/* Setup private info for the accelerator init */
		private->tiles_per_line = n_tiles;
		private->wide_tiles = wide_tile;
		/* No interleave : linear frame buffer
		** and a fairly medium fifo size */
		SETREG16(TILING_CONTROL, TWO_WAY_INTERLEAVE
			  | TILING_ENABLE
			  | (wide_tile ? WIDE_TILES : SMALL_TILES)
			  | ( (n_tiles << VIDEO_N_TILES_OFFSET)
				& VIDEO_N_TILES_MASK )
			  | ( 0x10 & FIFO_THRESHOLD_MASK )
			);
		/* Chaining */
		ORREG8(GR6_MISCELLANEOUS, GR6_GRAPHICS_MODE);
		/* Chaining again */
		ANDREG8(GR6_MISCELLANEOUS, ~GR6_CHAIN_ODD_TO_EVEN);
		/* Enable all planes 0 - 3 */
		SETREG8(SR2_PLANE_MASK, SR2_PLANE_0 | SR2_PLANE_1
			 | SR2_PLANE_2  | SR2_PLANE_3);
		/* Odd/Even settings */
		ORREG8(SR4_MEMORY_MODE, SR4_ODD_EVEN);
		/* CR17 here : various normalized settings */    
		SETREG8(CR17_MODE_CONTROL,
			CR17_ADDRESS_WRAP |
			CR17_BYTE_MODE |
			CR17_TIMING_ENABLE |
			CR17_CGA_SUPPORT |
			CR17_SELECT_ROW_SCAN_COUNTER
			);
		/* TEST (Be careful: add text mode change also)
		ANDREG8(SR4_MEMORY_MODE, ~SR4_CHAIN4);
		*/
		ANDREG8(GR5_MODE, ~GR5_ODD_EVEN);
		/* TEST (Be careful: add text mode change also)
		SETREG8(GR5_MODE, 0x42);
		*/
		/* Use ext memory */
		ORREG8(SR4_MEMORY_MODE, SR4_EXTENDED_MEMORY);

		/*
		ORREG8(GR5_MODE, GR5_256COLOR);
		*/

		/* Char cell height must be 1 ! */
		ANDREG8(CR9_CHARACTER_CELL_HEIGHT,
			~CR9_CELL_HEIGHT_MASK);
		ORREG8(SR1_CLOCKING_MODE, SR1_8_DOT_CLOCK);
		/* Setup doubleword mode */
		/*
		SETREG8(CR14_UNDERLINE_ROW_SCANLINE, CR14_DOUBLEWORD_MODE
			| CR14_COUNT_BY_FOUR );
			*/
		SETREG8(CR14_UNDERLINE_ROW_SCANLINE, 0);
		/* Origin is at 0,0 (0x0 in fb) */
		SETREG8(CRC_SCREEN_START_ADDRESS_HIGH, 0x00);
		SETREG8(CRD_SCREEN_START_ADDRESS_LOW, 0x00);
		ANDREG8(CR1B_EXTENDED_DISPLAY_CONTROL,
			~CR1B_SCREENA_ADDRESS_1718_MASK);
		ANDREG8(CR1D_SCREENA_START_EXTENSION,
			~CR1D_SCREENA_ADDRESS_1920_MASK);
		#warning also correct pixel panning and do a set origin -- ortalo
	}	/* else of if (mode->request.graphtype != KGIGT_TEXT16) */


	/*	Set the logical width, which spans three registers...
	*/
	SETREG8(CR13_OFFSET, LogicalWidth & 0xFF);
	/* EXTENDED REGISTERS */
	SETBITS8(CR1B_EXTENDED_DISPLAY_CONTROL,
		(LogicalWidth & 0x100) ? CR1B_OFFSET_8 : 0,
		CR1B_OFFSET_8);
	SETBITS8(CR1D_SCREENA_START_EXTENSION,
		(LogicalWidth & 0x200) ? CR1D_OFFSET_9 : 0,
		CR1D_OFFSET_9);

	/* Graphics Controller Bit Mask Register disabled
	*/
	SETREG8(GR8_BIT_MASK, 0xFF);

	/*	set palette
	*/
	for (i=0; i < 16; i++) {
		SETREG8(AR0_PALETTE+i, i);
	}
	/* set color select register */
	SETREG8(AR14_COLOR_SELECT, 0);

	/*
	**	setup horizontal timing
	*/
	NOTICE2("This is the mode we want to set:");
	#ifdef DEBUG
	TRACE1(printk_mode(mode,"mode of set_mode :"));
	#endif
	/*	horizontal display end
	*/ 
	i = (mode->x.width / DCLKdiv) - 1;
	SETREG8(CR1_HORIZONTAL_DISPLAY_END, i & 0xFF);
	/* EXTENDED REGISTERS */
	SETBITS8(CR1E_TIMING_OVERFLOW,
		(i & 0x100) ? CR1E_HOR_DISPLAY_END_8 : 0,
		CR1E_HOR_DISPLAY_END_8);
	/*	horizontal blank start
	*/
	DEBUG2("x.blankstart=%i", mode->x.blankstart);
	i = (mode->x.blankstart / DCLKdiv);
	SETREG8(CR2_HORIZONTAL_BLANKING_START, i & 0xFF);
	/* EXTENDED REGISTERS */
	SETBITS8(CR1E_TIMING_OVERFLOW,
		(i & 0x100) ? CR1E_HOR_BLANK_START_8 : 0,
		CR1E_HOR_BLANK_START_8);
	/*	horizontal sync start
	*/
	DEBUG2("x.syncstart=%i", mode->x.syncstart);
	i = (mode->x.syncstart / DCLKdiv);
	SETREG8(CR4_HORIZONTAL_SYNC_START, i & 0xFF);
	/* EXTENDED REGISTERS */
	SETBITS8(CR1E_TIMING_OVERFLOW,
		(i & 0x100) ? CR1E_HOR_SYNC_START_8 : 0,
		CR1E_HOR_SYNC_START_8);
	/*	horizontal sync end
	*/
	i = (mode->x.syncend / DCLKdiv);
	SETBITS8(CR5_HORIZONTAL_SYNC_END, i, CR5_HOR_SYNC_END_MASK);

	/*	horizontal blank end
	*/
	i = (mode->x.blankend / DCLKdiv);
	SETBITS8(CR3_HORIZONTAL_BLANKING_END, i, CR3_HOR_BLANK_END_MASK);
	SETBITS8(CR5_HORIZONTAL_SYNC_END, 
		(i & 0x20) ? CR5_HOR_BLANK_END_5 : 0,
		CR5_HOR_BLANK_END_5);
	/* EXTENDED REGISTERS */
	SETBITS8(CR1A_MISCELLANEOUS_CONTROL,
		((i & 0x40) ? CR1A_HOR_BLANK_END_6 : 0) |
		((i & 0x80) ? CR1A_HOR_BLANK_END_7 : 0),
		CR1A_HOR_BLANK_END_6 | CR1A_HOR_BLANK_END_7);
	/*	horizontal total
	*/
	DEBUG2("x.total=%i dclkdiv=%i", mode->x.total, DCLKdiv);
	i = (mode->x.total / DCLKdiv) - 5;	
	SETREG8(CR0_HORIZONTAL_TOTAL, i & 0xFF);
	/* EXTENDED REGISTERS */
	SETBITS8(CR1E_TIMING_OVERFLOW,
		(i & 0x100) ? CR1E_HOR_TOTAL_8 : 0,
		CR1E_HOR_TOTAL_8);
	/*
	**	setup vertical timing
	*/

	/*	vertical display end
	*/
	i = mode->y.width - 1;
	SETREG8(CR12_VERTICAL_DISPLAY_END, i & 0xFF);
	SETBITS8(CR7_OVERFLOW, 
		( (i & 0x100) ? CR7_VER_DISPLAY_END_8 : 0) |
		( (i & 0x200) ? CR7_VER_DISPLAY_END_9 : 0) ,
		CR7_VER_DISPLAY_END_8 | CR7_VER_DISPLAY_END_9);
	/* EXTENDED REGISTERS */
	SETBITS8(CR1E_TIMING_OVERFLOW,
		(i & 0x400) ? CR1E_VER_DISPLAY_END_10 : 0,
		CR1E_VER_DISPLAY_END_10);
	/*	vertical blank start
	*/
	i = mode->y.blankstart - 1;
	SETREG8(CR15_VERTICAL_BLANK_START, i & 0xFF);
	SETBITS8(CR7_OVERFLOW, 
		(i & 0x100) ? CR7_VER_BLANK_START_8 : 0,
		CR7_VER_BLANK_START_8);
	SETBITS8(CR9_CHARACTER_CELL_HEIGHT,
		(i & 0x200) ? CR9_VER_BLANK_START_9 : 0,
		CR9_VER_BLANK_START_9);
	/* EXTENDED REGISTERS */
	SETBITS8(CR1E_TIMING_OVERFLOW,
		(i & 0x400) ? CR1E_VER_BLANK_START_10 : 0,
		CR1E_VER_BLANK_START_10);
	/*	vertical sync start
	*/
	i = mode->y.syncstart;
	SETREG8(CR10_VERTICAL_SYNC_START, i & 0xFF);
	SETBITS8(CR7_OVERFLOW, 
		( (i & 0x100) ? CR7_VER_SYNC_START_8 : 0) |
		( (i & 0x200) ? CR7_VER_SYNC_START_9 : 0) ,
		CR7_VER_SYNC_START_8 | CR7_VER_SYNC_START_9);
	/* EXTENDED REGISTERS */
	SETBITS8(CR1E_TIMING_OVERFLOW,
		(i & 0x400) ? CR1E_VER_SYNC_START_10 : 0,
		CR1E_VER_SYNC_START_10);
	/*	vertical sync end
	*/
	i = mode->y.syncend;
	SETBITS8(CR11_VERTICAL_SYNC_END, i, CR11_VER_SYNC_END_MASK);

	/*	vertical blank end
	*/
	i = mode->y.blankend + 1;	
	SETREG8(CR16_VERTICAL_BLANK_END, i & 0xFF);
	/* EXTENDED REGISTERS */
	SETBITS8(CR1A_MISCELLANEOUS_CONTROL,
		((i & 0x100) ? CR1A_VER_BLANK_END_8 : 0) |
		((i & 0x200) ? CR1A_VER_BLANK_END_9 : 0),
		CR1A_VER_BLANK_END_8 | CR1A_VER_BLANK_END_9);
	/*	vertical total
	*/
	i = mode->y.total - 2;	
	SETREG8(CR6_VERTICAL_TOTAL, i & 0xFF);
	SETBITS8(CR7_OVERFLOW, 
		( (i & 0x100) ? CR7_VER_TOTAL_8 : 0) |
		( (i & 0x200) ? CR7_VER_TOTAL_9 : 0) ,
		CR7_VER_TOTAL_8 | CR7_VER_TOTAL_9);
	/* EXTENDED REGISTERS */
	SETBITS8(CR1E_TIMING_OVERFLOW,
		(i & 0x400) ? CR1E_VER_TOTAL_10 : 0,
		CR1E_VER_TOTAL_10);

#ifndef DIRECT_CARD_PROG
	DEBUG2("MISC before:%i", CLIO_GetStateRegister(&setup,MISC_OUTPUT,1));
#endif
	#warning we have a clock driver now SETBITS here -- ortalo
	SETREG8(MISC_OUTPUT, 
		MISC_ENABLE_DISPLAY_MEMORY |
		MISC_IOADDRESS_COLOR |
	        MISC_PAGE_SELECT | /* Idem VGA.c */
		((mode->x.polarity > 0) ? 0 : MISC_NEG_H_SYNC) |
		((mode->y.polarity > 0) ? 0 : MISC_NEG_V_SYNC));
#ifndef DIRECT_CARD_PROG
	DEBUG2("MISC after:%i", CLIO_GetStateRegister(&setup,MISC_OUTPUT,1));
#endif

	/*
	**	setup split line
	*/
	i = 0;
	SETREG8(CR18_LINE_COMPARE, i & 0xFF);
	SETBITS8(CR7_OVERFLOW, 
		(i & 0x100) ? CR7_LINE_COMPARE_8 : 0,
		CR7_LINE_COMPARE_8);
	SETBITS8(CR9_CHARACTER_CELL_HEIGHT,
		(i & 0x200) ? CR9_LINE_COMPARE_9 : 0,
		CR9_LINE_COMPARE_9);
	/* Enable hardware synchro on vertical blinking
	** for splitline updates (very convenient :-).
	*/
	ORREG8(CR1A_MISCELLANEOUS_CONTROL,
		CR1A_ENABLE_DOUBLE_BUFFERED_START_ADDRESS);

	/* Graphics Controller Bit Mask Register disabled
	*/
	SETREG8(GR8_BIT_MASK, 0xFF); /* TODO: Check why nothing former */
	/* set color select register */
	SETREG8(AR14_COLOR_SELECT, 0);

	/* Setup Rambus BCLK */
	/* WARNING: Too high clock may destroy your hardware ! */
	/* Here: 14.31818 MHz *	0x12 = 257.7 MHz (safe)
			(hmmm, the only one safe on my card...)
				0x13 = 272.0 MHz (pretty safe)
				0x14 = 286.4 MHz (fastest possible)
				0x15 = 300.7 MHz (5465 _only_!) */
	SETREG8(BCLK, 0x12 & BLCK_MULTIPLIER_MASK);

#ifndef DIRECT_CARD_PROG
	/*
	**	Ok. We'll do the programming now.
	*/
   	NOTICE2("We enter critical section"); 
	ENTER_CRITICAL;
#endif
	kgim_monitor_set_mode(dpy,mode);
	NOTICE2("monitor_set_mode in chipset:done");
	kgim_ramdac_set_mode(dpy,mode);
	NOTICE2("ramdac_set_mode in chipset:done");
	kgim_clock_set_mode(dpy,mode);
	NOTICE2("clock_set_mode in chipset:done");

#ifndef DIRECT_CARD_PROG
		RestoreState(&setup);
#endif
	/* Setting up the display struct now */
	/* Will also call kgim_accel_set_mode (if mode graphical) */
	chipset_setup_dpy(dpy, mode);

	LEAVE_CRITICAL;

	NOTICE2("Leaving chipset_set_mode");
	return;

	#undef TEXTX
	#undef TEXTY
	#undef ORREG8
	#undef ANDREG8
	#undef SETREG8
	#undef SETBITS8
}

kgi_sint kgim_chipset_command(struct kgi_graphic *state,kgi_uint cmd,kgi_uint arg)
{
	kgi_coord  *coord;
	struct kgi_display *dpy;

	dpy = state->dev.dpy;
	switch (cmd) {
#if 0
		case CHIP_SETSPLITLINE:
			if (arg > ( state->dev.mode.virt.y * state->dev.mode.dpp.y )
				return -E(CHIPSET,BOGUS);
			DEBUG1("IOCTL: set_split y=%d",arg);
			kgi_set_split(&(state->dev),arg*dpy->mode.magnify.y);
			return 0;
		case CHIP_SETVISFRAME:
			coord = (kgi_coord *)(state->io_buf);
			DEBUG1("IOCTL: set_origin x=%d y=%d virt.x=%d virt.y=%d",coord->x,coord->y,state->dev.mode.virt.x, state->dev.mode.virt.y);

			if (coord->x < 0 || 
			    coord->x > ( state->dev.mode.virt.x * state->dev.mode.dpp.x ))
				return -E(CHIPSET,BOGUS);

			if (coord->y < 0 || 
			    coord->y > ( state->dev.mode.virt.y * state->dev.mode.dpp.y ))
				return -E(CHIPSET,BOGUS);

			kgi_set_origin(&(state->dev), 0, coord->x, coord->y);
			return 0;
#endif
		default:
			DEBUG1("chipset_command: #%d (not implemented)\n",cmd);
			break;
	}

	return -E(CHIPSET,UNKNOWN);
}
